/* This file was hacked by solar to allow for some bad outputs from AVX2 */

/*
 * Copyright (c) 2013, Steve Thomas <steve AT tobtu DOT com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "bcrypt.h"
#include "bcryptavx2.h"
#include <stdio.h>
#include <string.h>
#ifdef _WIN32
	#include <windows.h>
	typedef LARGE_INTEGER     TIMER_TYPE;
	#define TIMER_FUNC(t)     QueryPerformanceCounter(&t)
	inline double TIMER_DIFF(LARGE_INTEGER s, LARGE_INTEGER e)
	{
		LARGE_INTEGER f;
		QueryPerformanceFrequency(&f);
		return ((double) (e.QuadPart - s.QuadPart)) / f.QuadPart;
	}

	#define WIN_ALIGN_32 __declspec(align(32))
	#define ALIGN_32
#else
	#include <sys/time.h>
	#define TIMER_TYPE        timeval
	#define TIMER_FUNC(t)     gettimeofday(&t, NULL)
	#define TIMER_DIFF(s,e)   ((e.tv_sec - s.tv_sec) + (e.tv_usec - s.tv_usec) / (double)1000000.0)

	#define WIN_ALIGN_32
	#define ALIGN_32 __attribute__((aligned(32)))
#endif

#define SWAP_ENDIAN_32(x) (((x) >> 24) | ((x) << 24) | (((x) >> 8) & 0xff00) | (((x) << 8) & 0xff0000))

const char *passwords[] = {"password", "123456", "qwerty", "Password1", "12345678", "monkey", "abc123", "111111"};
const char *base64Salt[] = {"xJaDaSyMzNxIsS9TGMb8R.", ".jcytsMoy8K6rxlGZZt9VO", "MtfvEju0JaMYSSqtbwN52e", "2yC6J7R0vYMrZsOn5EjTD.", "mYFMyQdonQCE56mOjsdAku", "VcoVe32XCDpU0aJBnO6so.", "TKrhpDQtWLhwTMlfvD4uG.", "gf8Jhqg9TcSFPd/Hm8b06O"};
const char *hashValues[] = {"5g8N4xikhDdXjaB7X2ynx8soi3XgRy.", "PWhm1oVp4UmFx2mB.LnDU1D/E1/tPii", "rlwOsVWKqsTIqlSI.p2PMSmpgJ7eju.", "jOYbRyieUlSL49yteFWvXbBurJI0mRq", "tyqhAXqGYbrlDRhV/eLJzEFnQB2S4iu", "L5kwlHojUTYtqjJmNA0SyIFCfDoxdQK", "ilehXcqcM//LlXMy/iPVlXSu5swHYI2", "kVBKnFhb1uyGwcOmx/X4mOeG4LJUhZm"};

int testAVX2(int verbose);
int testNormal(int verbose);
unsigned int benchAVX2  (const unsigned int *salt, unsigned int hash0, unsigned int hash1, unsigned int count, unsigned int rounds);
unsigned int benchNormal(const unsigned int *salt, unsigned int hash0, unsigned int hash1, unsigned int count, unsigned int rounds);

static int guessN = 8;

int testAVX2(int verbose)
{
	WIN_ALIGN_32
	unsigned int pw[8*18] ALIGN_32, salt[8*4] ALIGN_32, hash[8*6] ALIGN_32;
	unsigned char base64HashData[32], hashData[24];

	// Init test values
	for (int i = 0; i < 8; i++)
	{
		unsigned int tmp[4];

		base64Decode(tmp, base64Salt[i], 22);
		salt[        i] = SWAP_ENDIAN_32(tmp[0]);
		salt[1 * 8 + i] = SWAP_ENDIAN_32(tmp[1]);
		salt[2 * 8 + i] = SWAP_ENDIAN_32(tmp[2]);
		salt[3 * 8 + i] = SWAP_ENDIAN_32(tmp[3]);

		int len = strlen(passwords[i]) + 1;

		for (int j = 0; j < 18; j++)
		{
			pw[8 * j + i] = 
				(((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j    ) % len]) << 24) |
				(((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j + 1) % len]) << 16) |
				(((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j + 2) % len]) <<  8) |
				 ((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j + 3) % len]);
		}
	}

	// Run test values
	bcryptAVX2(pw, salt, hash, 32);

	// Check AVX2 test values
	int bad = 0;
	if (verbose)
	{
		printf("Checking AVX2 test values\n");
	}
	for (int i = 0; i < 8; i++)
	{
		for (int j = 0; j < 6; j++)
		{
			hashData[4 * j + 0] = (hash[8 * j + i] >> 24) & 0xff;
			hashData[4 * j + 1] = (hash[8 * j + i] >> 16) & 0xff;
			hashData[4 * j + 2] = (hash[8 * j + i] >>  8) & 0xff;
			hashData[4 * j + 3] =  hash[8 * j + i]        & 0xff;
		}
		base64Encode(base64HashData, hashData, 23);
		if (verbose)
		{
			printf("$2y$05$%s%s\n", base64Salt[i], base64HashData);
		}
		if (strcmp((char*) base64HashData, hashValues[i]))
		{
			if (guessN > i)
				guessN = i;
			bad++;
//			printf("Bad @%d\n", i);
		}
	}

	if (verbose)
	{
		printf("\nCorrect hashes:\n");
		for (int i = 0; i < 8; i++)
		{
			printf("$2y$05$%s%s\n", base64Salt[i], hashValues[i]);
		}
		if (bad == 0)
		{
			printf("\nAll hashes match.\n");
		}
		else
		{
			printf("\nERROR hashes do NOT match.\n");
		}
	}
	return bad;
}

int testNormal(int verbose)
{
	unsigned int pwU32[18*8], saltU32[4*8], hashU32[6*8];
	unsigned char base64HashData[32], hashData[24];

	// Init test values
	for (int i = 0; i < 8; i++)
	{
		unsigned int tmp[4];

		base64Decode(tmp, base64Salt[i], 22);
		((unsigned int*) saltU32)[4 * i    ] = SWAP_ENDIAN_32(tmp[0]);
		((unsigned int*) saltU32)[4 * i + 1] = SWAP_ENDIAN_32(tmp[1]);
		((unsigned int*) saltU32)[4 * i + 2] = SWAP_ENDIAN_32(tmp[2]);
		((unsigned int*) saltU32)[4 * i + 3] = SWAP_ENDIAN_32(tmp[3]);

		int len = strlen(passwords[i]) + 1;

		for (int j = 0; j < 18; j++)
		{
			pwU32[18 * i + j] = 
				(((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j    ) % len]) << 24) |
				(((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j + 1) % len]) << 16) |
				(((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j + 2) % len]) <<  8) |
				 ((unsigned int) ((const unsigned char*) (passwords[i]))[(4 * j + 3) % len]);
		}
	}

	// Run test values
	for (int i = 0; i < 8; i++)
	{
		bcrypt(pwU32 + 18 * i, saltU32 + 4 * i, hashU32 + 6 * i, 32);
	}

	// Check normal test values
	int bad = 0;
	if (verbose)
	{
		printf("Checking normal test values\n");
	}
	for (int i = 0; i < 8; i++)
	{
		for (int j = 0; j < 6; j++)
		{
			hashData[4 * j + 0] = (hashU32[6 * i + j] >> 24) & 0xff;
			hashData[4 * j + 1] = (hashU32[6 * i + j] >> 16) & 0xff;
			hashData[4 * j + 2] = (hashU32[6 * i + j] >>  8) & 0xff;
			hashData[4 * j + 3] =  hashU32[6 * i + j]        & 0xff;
		}
		base64Encode(base64HashData, hashData, 23);
		if (verbose)
		{
			printf("$2y$05$%s%s\n", base64Salt[i], base64HashData);
		}
		if (strcmp((char*) base64HashData, hashValues[i]))
		{
			bad++;
		}
	}

	if (verbose)
	{
		printf("\nCorrect hashes:\n");
		for (int i = 0; i < 8; i++)
		{
			printf("$2y$05$%s%s\n", base64Salt[i], hashValues[i]);
		}
		if (bad == 0)
		{
			printf("\nAll hashes match.\n");
		}
		else
		{
			printf("\nERROR hashes do NOT match.\n");
		}
	}
	return bad;
}

unsigned int benchAVX2(const unsigned int *salt, unsigned int hash0, unsigned int hash1, unsigned int count, unsigned int rounds)
{
	WIN_ALIGN_32
	unsigned int pwAVX2[8*18] ALIGN_32, saltAVX2[8*4] ALIGN_32, hashAVX2[8*2] ALIGN_32;
	unsigned int found = 0;

	// Init salt
	for (unsigned int i = 0; i < 4*8; i += 4)
	{
		saltAVX2[i    ] = salt[0];
		saltAVX2[i + 1] = salt[1];
		saltAVX2[i + 2] = salt[2];
		saltAVX2[i + 3] = salt[3];
	}

	// Init PW
	for (unsigned int i = 0; i < 8; i++)
	{
		for (unsigned int j = 0; j < 18; j++)
		{
			pwAVX2[18 * i + j] = 8 * count + i;
		}
	}

	// Bench
	for (unsigned int i = 0; i < count; i++)
	{
		found += bcryptAVX2(pwAVX2, saltAVX2, hashAVX2, rounds, true, hash0, hash1);

		// Increment PW
		for (unsigned int i = 0; i < 8; i++)
		{
			for (unsigned int j = 0; j < 18; j += 6)
			{
				pwAVX2[18 * i + j    ] += 8;
				pwAVX2[18 * i + j + 1] += 8;
				pwAVX2[18 * i + j + 2] += 8;
				pwAVX2[18 * i + j + 3] += 8;
				pwAVX2[18 * i + j + 4] += 8;
				pwAVX2[18 * i + j + 5] += 8;
			}
		}
	}
	return found;
}

unsigned int benchNormal(const unsigned int *salt, unsigned int hash0, unsigned int hash1, unsigned int count, unsigned int rounds)
{
	unsigned int pwU32[18], saltU32[4], hashU32[2];
	unsigned int found = 0;

	// Init salt
	saltU32[0] = salt[0];
	saltU32[1] = salt[1];
	saltU32[2] = salt[2];
	saltU32[3] = salt[3];

	// Init PW
	for (unsigned int i = 0; i < 18; i++)
	{
		pwU32[i] = count;
	}

	// Bench
	for (unsigned int i = 0; i < count; i++)
	{
		found += bcrypt(pwU32, saltU32, hashU32, rounds, true, hash0, hash1);

		// Increment PW
		for (unsigned int i = 0; i < 18; i += 6)
		{
			pwU32[i    ]++;
			pwU32[i + 1]++;
			pwU32[i + 2]++;
			pwU32[i + 3]++;
			pwU32[i + 4]++;
			pwU32[i + 5]++;
		}
	}
	return found;
}

int main()
{
	int badAVX2 = 0, badNormal = 0;

	badAVX2 = testAVX2(0);
	badNormal = testNormal(0);
	if (badAVX2 + badNormal == 0)
	{
		printf("Tests PASSED\n");
	}
	else
	{
		if (guessN && !badNormal)
			printf("Some tests failed, but this may be OK since %d AVX2 code outputs are correct\n", guessN);
		else
			printf("Tests FAILED\n");
	}
	if (badAVX2)
	{
		printf("AVX2 bcrypt failed to produce valid hashes.\n");
	}
	if (badNormal)
	{
		printf("Normal bcrypt failed to produce valid hashes.\n");
	}

	printf("Benchmarking...\n");
	unsigned int salt[] = {0x00112233, 0x44556677, 0x8899aabb, 0xccddeeff};
	unsigned int hash0 = 0;
	unsigned int hash1 = 0;
	unsigned int rounds = 32;

	TIMER_TYPE s, e;
	unsigned int count = 1;
	unsigned int dummy = 0;
	do
	{
		count <<= 1;
		TIMER_FUNC(s);
		dummy += benchAVX2(salt, hash0, hash1, count, rounds);
		TIMER_FUNC(e);
	} while (TIMER_DIFF(s, e) < 1);
	printf("AVX2: %d*%u took %0.4f sec (%0.1f h/s) %u\n", guessN, count, TIMER_DIFF(s, e), guessN * count / TIMER_DIFF(s, e), dummy);

	count = 1;
	dummy = 0;
	do
	{
		count <<= 1;
		TIMER_FUNC(s);
		dummy += benchNormal(salt, hash0, hash1, count, rounds);
		TIMER_FUNC(e);
	} while (TIMER_DIFF(s, e) < 1);
	printf("Normal: %u took %0.4f sec (%0.1f h/s) %u\n", count, TIMER_DIFF(s, e), count / TIMER_DIFF(s, e), dummy);
	return 0;
}
